﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.IO;
using System.Windows.Forms;
using ChaKi.Entity.Corpora;
using ChaKi.Entity.Corpora.Annotations;
using ChaKi.Service.DependencyEdit;

namespace DependencyEditSLA
{
    enum DADragType
    {
        None = 0,
        ArrowEnd
    }

    public partial class SentenceStructure : UserControl
    {
        #region private fields
        private Sentence m_Model;     // Model object
        private DepEditServiceSLA m_Service;    // Service object

        private List<Segment> m_Segments;
        private List<Link> m_Links;

        private Word m_CenterWord;
        private List<BunsetsuBox> m_BunsetsuBoxes;
        private List<ConcatButton> m_ConcatButtons;
        private List<DepArrow> m_Arrows;

        private DepArrow m_SelArrow;    // 選択中の矢印（null可）
        private DepArrow m_DragArrow;   // ドラッグ中の矢印（null可）
        private DADragType m_DragType;  // ドラッグ種別
        private BunsetsuBox m_OnBunsetsuBox;    // ドラッグ中にマウスのかかった文節Box（強調表示する; null可）
        #endregion

        public event EventHandler OnLayoutChanging;
        public event EventHandler OnLayoutChanged;

        /// <summary>
        /// コンストラクタ
        /// </summary>
        public SentenceStructure()
        {
            InitializeComponent();

            m_Model = null;
            m_Service = null;
            m_BunsetsuBoxes = new List<BunsetsuBox>();
            m_ConcatButtons = new List<ConcatButton>();
            m_Arrows = new List<DepArrow>();
            m_CenterWord = null;
            m_SelArrow = null;
            m_OnBunsetsuBox = null;
        }

        /// <summary>
        /// このコントロールの表示の元となるSentence（モデルオブジェクト）を設定する
        /// </summary>
        /// <param name="sen"></param>
        public void SetSentence(Sentence sen, DepEditServiceSLA svc)
        {
            m_Model = sen;
            m_Service = svc;
        }

        /// <summary>
        /// 指定したWordの強調表示を指示する.
        /// </summary>
        /// <param name="word"></param>
        public void SetCenterWord(int cid)
        {
            if (m_Model != null && cid >= 0 && cid < m_Model.Words.Count)
            {
                m_CenterWord = (Word)m_Model.Words[cid];
            }
        }

        /// <summary>
        /// モデルを元にコントロールを作る。
        /// 既にコントロールが存在した場合は、消去して作り直す。
        /// </summary>
        public void UpdateContents()
        {
            m_BunsetsuBoxes.Clear();
            m_ConcatButtons.Clear();
            m_Arrows.Clear();
            this.Controls.Clear();
            this.Height = 0;
            this.Width = 0;

            this.Visible = false;
            this.SuspendLayout();

            m_Service.GetBunsetsuTags(out m_Segments, out m_Links);
            if (m_Segments.Count == 0)
            {
                return;
            }

            // 文節に対応するBunsetsuBoxを作成する
            for (int i = 0; i < m_Segments.Count; i++)
            {
                Segment seg = m_Segments[i];
                // 文節Box
                BunsetsuBox bbox = new BunsetsuBox(seg);
                m_BunsetsuBoxes.Add(bbox);
                this.Controls.Add(bbox);
                bbox.OnSplit += new MergeSplitEventHandler(this.OnSplitBunsetsu);
                // 結合ボタン
                if (i < m_Segments.Count - 1)
                {
                    ConcatButton cb = new ConcatButton(i);
                    m_ConcatButtons.Add(cb);
                    this.Controls.Add(cb);
                    cb.Click += new System.EventHandler(this.OnConcatBunsetsu);
                }
            }
            Debug.Assert(m_BunsetsuBoxes.Count > 0);

            // それぞれのWordを文節に割り当てていく
            foreach (Segment b in m_Segments)
            {
                foreach (Word w in m_Model.Words)
                {
                    if (w.StartChar >= b.StartChar && w.StartChar < b.EndChar)
                    {
                        BunsetsuBox bbx = FindBunsetsuBox(b);
                        if (bbx != null)
                        {
                            WordBox wb = bbx.AddWordBox(w, (w == m_CenterWord));
                            wb.OnChagneLexeme += new ChangeLexemeEventHandler(this.OnChangeLexeme);
                        }
                    }
                }
            }

            this.RecalcLayout();

            // 依存矢印(DepArrow)を作成する
            for (int i = 0; i < m_Links.Count; i++)
            {
                Link link = m_Links[i];
                Segment b1 = link.From;
                if (b1 == null) continue;
                Segment b2 = link.To;
                if (b2 == null) continue;
                BunsetsuBox box1 = FindBunsetsuBox(b1);
                int box1index = m_BunsetsuBoxes.IndexOf(box1);
                BunsetsuBox box2 = FindBunsetsuBox(b2);
                int box2index = m_BunsetsuBoxes.IndexOf(box2);
                if (box1 != null && box2 != null)
                {
                    DepArrow dp = new DepArrowArc(link, i, box1index, box2index, box1.Bounds, box2.Bounds);
                    m_Arrows.Add(dp);
                    dp.TagLabel.TagChanged += new EventHandler(this.OnTagChanged);
                    this.Controls.Add(dp.TagLabel);
                }
            }
            CheckCrossDep();
            this.ResumeLayout();
            this.Visible = true;
//            this.Scale(new SizeF(0.5F, 0.5F));
//            ScrollControlIntoView(m_BunsetsuBoxes[0]);
        }

        /// <summary>
        /// コントロールの配置を調整する
        /// </summary>
        public void RecalcLayout()
        {
            if (OnLayoutChanging != null) OnLayoutChanging(this,null);

            int x = 10;
            int height = 0;
            // 全体の幅を計算・再配置し、それに基づいて必要高さを計算する
            for (int i = 0; i < m_BunsetsuBoxes.Count; i++)
            {
                BunsetsuBox bb = m_BunsetsuBoxes[i];
                bb.RecalcLayout();
                int width = bb.Width;
                bb.Left = x;
                x += (width + 4);
                if (i < m_BunsetsuBoxes.Count - 1)
                {
                    ConcatButton cb = m_ConcatButtons[i];
                    cb.Left = x;
                    x += cb.Width;
                }
                x += 4;
                height = Math.Max(height, bb.Height);
            }
            int y = height + (int)((x - 10) * DrawParamsArc.CurveParamY);

            // 全体のY位置を下詰めで計算する
            foreach (BunsetsuBox bb in m_BunsetsuBoxes)
            {
                bb.Top = y - height;
            }
            foreach (ConcatButton cb in m_ConcatButtons)
            {
                cb.Top = y - 22;
            }
            this.Width = x;
            this.Height = y;
            if (OnLayoutChanged != null) OnLayoutChanged(this,null);
        }

        /// <summary>
        /// 文節オブジェクトからそのViewとなっているBoxを得る
        /// </summary>
        /// <param name="b"></param>
        /// <returns></returns>
        private BunsetsuBox FindBunsetsuBox(Segment b)
        {
            foreach (BunsetsuBox bb in m_BunsetsuBoxes)
            {
                if (bb.Model == b)
                {
                    return bb;
                }
            }
            return m_BunsetsuBoxes[0];
        }

        /// <summary>
        /// 依存矢印の交差がないかどうか判定する
        /// </summary>
        /// <returns></returns>
        public bool CheckCrossDep()
        {
            bool has_cross = false;
            foreach (DepArrow da in m_Arrows)
            {
                da.IsCrossing = false;
            }
            if (m_Arrows.Count < 2)
            {
                return false;
            }
            foreach (DepArrow da1 in m_Arrows)
            {
                int da1_s = da1.FromIndex;
                int da1_e = da1.ToIndex;
                int da1_min = Math.Min(da1_s, da1_e);
                int da1_max = Math.Max(da1_s, da1_e);
                foreach (DepArrow da2 in m_Arrows)
                {
                    int da2_s = da2.FromIndex;
                    int da2_e = da2.ToIndex;
                    int da2_min = Math.Min(da2_s, da2_e);
                    int da2_max = Math.Max(da2_s, da2_e);
                    if ((da1_min <= da2_min && da1_max >= da2_max)
                     || (da1_min >= da2_min && da1_max <= da2_max)
                     || (Math.Max(da1_s, da1_e) <= Math.Min(da2_s, da2_e))
                     || (Math.Max(da2_s, da2_e) <= Math.Min(da1_s, da1_e)))
                    {
                        continue;
                    }
                    else
                    {
                        da1.IsCrossing = true;
                        da2.IsCrossing = true;
                        has_cross = true;
                    }
                }
            }
            return has_cross;
        }

        /// <summary>
        /// 文節の切断が指示された場合のハンドラ
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void OnSplitBunsetsu(object sender, MergeSplitEventArgs e)
        {
            // Modelに対して文節切断を行う。
            try
            {
                m_Service.SplitBunsetsu(e.StartPos, e.EndPos, e.SplitPos);
            }
            catch (Exception ex)
            {
                ErrorReportDialog dlg = new ErrorReportDialog("Cannot Execute operation", ex);
                dlg.ShowDialog();
            }
            UpdateContents();
        }

        /// <summary>
        /// 語のLexeme対応変更が指示された場合のハンドラ
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void OnChangeLexeme(object sender, ChangeLexemeEventArgs e)
        {
            try
            {
                m_Service.ChangeLexeme(e.wordPos, e.newLex);
            }
            catch (Exception ex)
            {
                ErrorReportDialog dlg = new ErrorReportDialog("Cannot Execute operation", ex);
                dlg.ShowDialog();
            }
            UpdateContents();
        }

        /// <summary>
        /// 「＋」ボタンが押されたときの処理
        /// 文節をマージする。
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void OnConcatBunsetsu(object sender, EventArgs e)
        {
            //TODO: 文節間に直接の依存関係がなければ警告を出す。

            try
            {
                int b_pos = ((ConcatButton)sender).Pos;  // 「＋」ボタンの左にある文節のPos
                if (b_pos >= m_Segments.Count-1)
                {
                    return;
                }
                Segment left = m_Segments[b_pos];
                Segment right = m_Segments[b_pos + 1];

                m_Service.MergeBunsetsu(left.StartChar, right.EndChar, left.EndChar);
            }
            catch (Exception ex)
            {
                ErrorReportDialog dlg = new ErrorReportDialog("Cannot Execute operation", ex);
                dlg.ShowDialog();
            }
            UpdateContents();
        }

        private void OnTagChanged(object sender, EventArgs e)
        {
            TagLabel label = (TagLabel)sender;      //TODO: Labelではなく、Tag自体の変更にしなければならない
            Link link = label.Link;
            try
            {
                string curtag = link.Tag.Name;
                string newtag = label.Text;
                m_Service.ChangeLinkTag(link, curtag, newtag);
            }
            catch (Exception ex)
            {
                MessageBox.Show("Operation Failed:" + ex);
            }
            UpdateContents();
        }

        /// <summary>
        /// 追加の描画処理
        /// 文節Box間に依存矢印を描画する。
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void SentenceStructure_Paint(object sender, PaintEventArgs e)
        {
            e.Graphics.SmoothingMode = SmoothingMode.AntiAlias;
            e.Graphics.TextRenderingHint = System.Drawing.Text.TextRenderingHint.AntiAlias;

            foreach (DepArrow arrow in m_Arrows)
            {
                arrow.Draw(e.Graphics, (m_DragType != DADragType.None), this.DisplayRectangle.Location);
            }
        }

        /// <summary>
        /// マウス左押下時の処理
        /// DepArrow に対してヒットテストを行う。
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void SentenceStructure_MouseDown(object sender, MouseEventArgs e)
        {
            if (e.Button != MouseButtons.Left)
            {
                return;
            }

            // 矢印のヒットテスト
            DAHitType ht;
            UnpickDepArrow();
            m_DragArrow = PickDepArrow(e.Location, out ht);
            if (m_DragArrow != null)
            {
                m_DragArrow.IsSelected = true;
                if (ht == DAHitType.Other)
                {
                    m_DragType = DADragType.ArrowEnd;
                    this.Capture = true;
                }
                else if (ht == DAHitType.TagText)
                {
                    m_SelArrow = m_DragArrow;
                    m_DragArrow = null;
                }
            }
        }

        private void SentenceStructure_MouseMove(object sender, MouseEventArgs e)
        {
            if (m_DragType == DADragType.ArrowEnd)
            {
                if (m_DragArrow != null)
                {
                    // マウスが左右境界を越えたらスクロールを行う。
                    if (e.X > this.Width)
                    {
                        this.HorizontalScroll.Value = Math.Min( this.HorizontalScroll.Value+10, this.HorizontalScroll.Maximum );
                    }
                    else if (e.X < 0)
                    {
                       this.HorizontalScroll.Value = Math.Max(this.HorizontalScroll.Value-10, this.HorizontalScroll.Minimum);
                    }
                    // 矢印線の描画更新
                    this.Invalidate(m_DragArrow.Rgn);
                    Point newpnt = new Point( e.Location.X - this.DisplayRectangle.Left, e.Location.Y - this.DisplayRectangle.Top );
                    m_DragArrow.MoveHeadTo(newpnt);
                    this.Invalidate(m_DragArrow.Rgn);

                    //  Phraseの強調
                    BunsetsuBox pb = HitTestBunsetsuBox(e.Location);
                    if (pb != null)
                    {
                        if (pb != m_OnBunsetsuBox)
                        {
                            if (m_OnBunsetsuBox != null)
                            {
                                m_OnBunsetsuBox.Hover = false;
                            }
                            m_OnBunsetsuBox = pb;
                            m_OnBunsetsuBox.Hover = true;
                            m_OnBunsetsuBox.Invalidate();
                            Application.DoEvents();
                        }
                    }
                    else
                    {
                        if (m_OnBunsetsuBox != null)
                        {
                            m_OnBunsetsuBox.Hover = false;
                            m_OnBunsetsuBox.Invalidate();
                            Application.DoEvents();
                        }
                        m_OnBunsetsuBox = null;
                    }
                }
            }
        }

        private void SentenceStructure_MouseUp(object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Left)
            {
                if (m_OnBunsetsuBox != null)
                {
                    m_OnBunsetsuBox.Hover = false;
                }
                if (m_DragType == DADragType.ArrowEnd && m_DragArrow != null)
                {
                    this.Capture = false;
                    // 後でUpdateContentsした後にスクロール位置を戻すため、現在のスクロール位置を記憶する
                    Point offset0 = this.DisplayRectangle.Location;

                    // 終点が文節Box内かどうか判定
                    BunsetsuBox pb = HitTestBunsetsuBox(e.Location);
                    Link link = m_DragArrow.Link;
                    if (pb != null && pb.Model != link.From)
                    {
                        try
                        {
                            m_Service.ChangeLinkEnd(link, link.To, pb.Model);
                        }
                        catch (Exception ex)
                        {
                            ErrorReportDialog dlg = new ErrorReportDialog("Cannot Execute operation", ex);
                            dlg.ShowDialog();
                        }
                        UpdateContents();
                    }
                    else
                    {
                        UpdateContents();
                    }
                    SetDisplayRectLocation(offset0.X, offset0.Y);
                    AdjustFormScrollbars(true);
                    Invalidate();
                }
                m_DragType = DADragType.None;
                m_DragArrow = null;
            }
        }

        private DepArrow HitTestDepArrow(Point pt, out DAHitType ht)
        {
            Graphics g = this.CreateGraphics();
            foreach (DepArrow arrow in m_Arrows)
            {
                ht = arrow.HitTest(g, pt);
                if (ht != DAHitType.None)
                {
                    return arrow;
                }
            }
            g.Dispose();
            ht = DAHitType.None;
            return null;
        }

        private BunsetsuBox HitTestBunsetsuBox(Point pt)
        {
            foreach (BunsetsuBox pb in m_BunsetsuBoxes)
            {
                if (pb.Bounds.Contains(pt))
                {
                    return pb;
                }
            }
            return null;
        }

        private DepArrow PickDepArrow(Point pt, out DAHitType ht)
        {
            DepArrow arrow = HitTestDepArrow(pt, out ht);
            if (arrow != null)
            {
                m_SelArrow = arrow;
                return arrow;
            }
            return null;
        }

        private void UnpickDepArrow()
        {
            foreach (DepArrow arrow in m_Arrows)
            {
                arrow.IsSelected = false;
            }
            m_SelArrow = null;
        }

        public void Undo()
        {
            try
            {
                if (m_Service.Undo())
                {
                    UpdateContents();
                }
            }
            catch (Exception ex)
            {
                ErrorReportDialog dlg = new ErrorReportDialog("Cannot Execute operation", ex);
                dlg.ShowDialog();
            }
        }

        public void Redo()
        {
            try
            {
                if (m_Service.Redo())
                {
                    UpdateContents();
                }
            }
            catch (Exception ex)
            {
                ErrorReportDialog dlg = new ErrorReportDialog("Cannot Execute operation", ex);
                dlg.ShowDialog();
            }
        }

        public bool CanUndo()
        {
            if (m_Service != null)
            {
                return m_Service.CanUndo();
            }
            return false;
        }

        public bool CanRedo()
        {
            if (m_Service != null)
            {
                return m_Service.CanRedo();
            }
            return false;
        }

        public bool CanSave()
        {
            if (m_Service != null)
            {
                return m_Service.CanSave();
            }
            return false;
        }

        public void WriteToDotFile(string filename)
        {
            if (m_Model == null)
            {
                return;
            }
            TextWriter wr = null;
            try
            {
                wr = new StreamWriter(filename);
                m_Service.WriteToDotFile(wr);
            }
            catch (IOException ex)
            {
                ErrorReportDialog dlg = new ErrorReportDialog("Cannot Execute operation", ex);
                dlg.ShowDialog();
            }
            finally
            {
                if (wr != null)
                {
                    wr.Close();
                }
            }
        }
    }
}
